/**
 * PETALS - PETALS Services Platform. Copyright (c) 2007 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */
package org.ow2.petals.bc.ftp.service;

import java.io.IOException;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.logging.Logger;

import javax.activation.DataHandler;
import javax.jbi.JBIException;
import javax.xml.transform.Source;
import javax.xml.transform.stream.StreamSource;

import org.ow2.petals.bc.ftp.FTPConstants;
import org.ow2.petals.bc.ftp.FTPUtil;
import org.ow2.petals.bc.ftp.connection.FTPConnection;
import org.ow2.petals.bc.ftp.connection.FTPConnectionInfo;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.util.XMLUtil;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * The FTP JBI listener. Writes a file to FTP server on incoming JBI message.
 * 
 * @author alouis
 */
public class FTPService {

    private Logger logger;

    public FTPService(Logger logger) {
        super();
        this.logger = logger;
    }

    /**
     * Put the source on the ftp server with the given filename a Fault is
     * returned on IO problems TODO the content to send is a String, which is
     * not optimal
     * 
     * @param exchange
     * @param fileName
     * @param connectionInfo
     * @param inMessage 
     * @throws JBIException
     *             bad MEP
     */
    public void processPut(final Exchange exchange, final FTPConnectionInfo connectionInfo, 
            final Document inMessage)
            throws JBIException {
        if (exchange == null)
            throw new NullPointerException("exchange parameter is null");

        logger.info("process PUT operation");
        if (!exchange.isInOutPattern()) {

            try {
                Node root = inMessage.getFirstChild();
                if (root == null || !FTPConstants.PUT.equalsIgnoreCase(root.getLocalName())) {
                    FTPUtil.setMissingElementFaultOnExchange(exchange, FTPConstants.PUT);
                } else {
                    Node fileNameNode = XMLUtil.findChild(root, FTPConstants.FTP_FILENAME, null,
                            false);
                    if (fileNameNode == null
                            || !FTPConstants.FTP_FILENAME.equalsIgnoreCase(fileNameNode
                                    .getLocalName())) {
                        FTPUtil.setMissingElementFaultOnExchange(exchange,
                                FTPConstants.FTP_FILENAME);
                    } else {
                        Node contentNode = XMLUtil.findChild(root, FTPConstants.BODY, null, false);
                        if (contentNode == null
                                || !FTPConstants.BODY.equalsIgnoreCase(contentNode.getLocalName())) {
                            FTPUtil.setMissingElementFaultOnExchange(exchange, FTPConstants.BODY);
                        } else {
                            String fileName = fileNameNode.getTextContent();
                            String content = XMLUtil.createStringFromDOMNodeList(contentNode.getChildNodes());
                            FTPConnection ftpConnection = new FTPConnection(connectionInfo, logger);
                            ftpConnection.putString(fileName, content);
                        }
                    }
                }
            } catch (Exception e) {
                FTPUtil.setIOFaultOnExchange(exchange, "Can not put XML content on FTP server ["
                        + connectionInfo.getServer() + "]. Cause : " + e.getMessage());
            }
        } else {
            throw new JBIException("Exchange pattern for PUT operation can not be In-Out");
        }
        logger.finest("PUT operation ended");
    }

    /**
     * Put the body element of the XML message on the ftp server. a Fault is
     * returned on IO problems TODO the content to send is a String, which is
     * not optimal
     * 
     * @param exchange
     * @param fileName
     * @param connectionInfo
     * @param inMessage
     * @throws JBIException
     *             bad MEP
     */
    public void processStandardPut(final Exchange exchange, String fileName,
            final FTPConnectionInfo connectionInfo, final Document inMessage) throws JBIException {
        if (exchange == null)
            throw new NullPointerException("exchange parameter is null");

        logger.info("process standard PUT operation (the whole XML message put on the FTP server)");
        if (!exchange.isInOutPattern()) {

            try {
                String content = XMLUtil.createStringFromDOMNode(inMessage);
                FTPConnection ftpConnection = new FTPConnection(connectionInfo, logger);
                ftpConnection.putString(fileName, content);

            } catch (Exception e) {
                FTPUtil.setIOFaultOnExchange(exchange, "Can not put XML content on FTP server ["
                        + connectionInfo.getServer() + "]. Cause : " + e.getMessage());
            }
        } else {
            throw new JBIException("Exchange pattern for standard PUT operation can not be In-Out");
        }
        logger.finest("PUT operation ended");
    }

    /**
     * Put attachments on the ftp server TODO an FTP connection is done for each
     * file...
     * 
     * @param exchange
     * @param connectionInfo
     * @throws JBIException
     */
    public void processMPut(final Exchange exchange, final FTPConnectionInfo connectionInfo)
            throws JBIException {
        logger.info("process MPUT operation");
        if (!exchange.isInOutPattern()) {
            try {
                FTPConnection ftpConnection = new FTPConnection(connectionInfo, logger);
                
                // rebuild a hashmap containing files and their names
                HashMap<String, DataHandler> attachments = new HashMap<String, DataHandler>();
                for (Object fileName : exchange.getInMessage().getAttachmentNames()) {
                    attachments.put((String)fileName,exchange.getInMessage().getAttachment((String)fileName));
                }
                ftpConnection.mput(attachments);
            } catch (Exception e) {
                if (!exchange.isInOnlyPattern()) {
                    FTPUtil.setIOFaultOnExchange(exchange, "Can not copy file(s) on FTP server ["
                            + connectionInfo.getServer() + "]. Cause : " + e.getMessage());
                }
            }
        } else {
            throw new JBIException("Exchange pattern for PUT operation can not be In-Out");
        }
        logger.finest("MPUT operation ended");
    }

    /**
     * FTP DIR operation. <br>
     * a Fault is returned for bad request or IO problems
     * 
     * @param exchange
     *            the message exchange, MUST be non null
     * @throws JBIException
     *             bad MEP
     */

    public void processDir(Exchange exchange, final FTPConnectionInfo connectionInfo)
            throws JBIException {
        logger.info("process DIR operation");

        if (exchange.isInOutPattern() || exchange.isInOptionalOutPattern()) {

            List<String> fileNames;
            try {
                // do the FTP DIR operation
                FTPConnection ftpConnection = new FTPConnection(connectionInfo, logger);
                fileNames = ftpConnection.list();

                // generate filename list
                String result = FTPUtil.generateFileNameList(fileNames, "dir");
                exchange.setOutMessageContent(result);
            } catch (IOException e) {
                FTPUtil.setIOFaultOnExchange(exchange, "Can not list files on FTP server ["
                        + connectionInfo.getServer() + "]. Cause : " + e.getMessage());
            }

        } else {
            throw new JBIException(
                    "Exchange pattern for DIR operation must be IN-OUT or IN-OPTIONAL-OUT");
        }
        logger.finest("DIR operation ended");
    }

    /**
     * FTP MGET operation <br>
     * All files are getted as attachments
     * 
     * @param exchange
     * @param connectionInfo
     * @param inMessage
     * @throws Exception
     */
    public void processMGet(Exchange exchange, final FTPConnectionInfo connectionInfo ,
            final Document inMessage) throws Exception {

        logger.info("process MGET operation");
        if (exchange.isInOutPattern() || exchange.isInOptionalOutPattern()) {

            // retrieve filenames to get
            NodeList list = inMessage.getFirstChild().getChildNodes();
            List<String> filenames = new LinkedList<String>();
            for (int index = 0; index < list.getLength(); index++) {
                if (FTPConstants.FTP_FILENAME.equalsIgnoreCase(list.item(index).getLocalName())) {
                    filenames.add(list.item(index).getTextContent());
                }
            }
            // do the FTP MGET operation
            FTPConnection ftpConnection = new FTPConnection(connectionInfo, logger);
            Map<String, DataHandler> files = ftpConnection.mGet(filenames);

            // add attachments on the OUT message
            for (String filename : files.keySet()) {
                exchange.getOutMessage().addAttachment(filename, files.get(filename));
            }
            List<String> retrievedFilenames = new LinkedList<String>(files.keySet());
            String response = FTPUtil.generateFileNameList(retrievedFilenames, "mget");
            exchange.setOutMessageContent(response);
        } else {
            throw new Exception(
                    "Exchange pattern for MGET operation MUST be IN-OUT or IN-OPTIONAL-OUT");
        }
        logger.finest("MGET operation ended");
    }

    /**
     * 
     * FTP GET operation <br>
     * get the file as a source message a Fault is returned for bad request or
     * IO problems
     * 
     * @param exchange
     *            the message exchange, MUST be non null
     * @param connectionInfo
     * @param inMessage
     * @throws JBIException
     *             bad MEP
     */
    public void processGet(final Exchange exchange, final FTPConnectionInfo connectionInfo,
            final Document inMessage)
            throws JBIException {
        logger.info("process GET operation");

        // get accepts only InOut and InOptionalOut message patterns
        if (exchange.isInOptionalOutPattern() || exchange.isInOutPattern()) {

            // retrieve the filename to get
            Node filePatternNode = XMLUtil.findChild(inMessage.getFirstChild(), FTPConstants.FTP_FILENAME,
                    null, false);
            String filePattern = null;
            if (filePatternNode != null) {
                filePattern = filePatternNode.getTextContent();
            }

            if (filePattern != null) {
                try {
                    // do the FTP GET operation
                    FTPConnection ftpConnection = new FTPConnection(connectionInfo, logger);
                    DataHandler dh = ftpConnection.get(filePattern);
                    Source stream = new StreamSource(dh.getInputStream());

                    exchange.setOutMessageContent(stream);
                } catch (IOException e) {
                    FTPUtil.setIOFaultOnExchange(exchange, "Can not get file [" + filePattern
                            + "] on FTP server [" + connectionInfo.getServer() + "]. Cause : "
                            + e.getMessage());
                }
            } else {
                FTPUtil.setMissingElementFaultOnExchange(exchange, "filename");
            }
        } else {
            throw new JBIException(
                    "Exchange pattern for GET operation MUST be IN-OUT or OPTIONAL-IN-OUT");
        }
        logger.finest("GET operation ended");
    }
}
